{#
Renders a basic plot div, that can be used in conjunction with PLOT_JS.

:param elementid: a unique identifier for the ``<div>`` a PLOT_JS
    template should be configured with the same ``elementid``
:type elementid: str

#}
<div class="bk-root">
    <div class="bk-plotdiv" id="{{ elementid }}"></div>
</div>
